module TrRMIte
  module Consumers
    class Scheduler < Consumer

      include EventHandling::DSL
      include EventHandling::Processing

      def self.default_name
        'EVENT_SCHEDULER'
      end

      def initialize(topic, registry_uri, queue_uri, interval, state_store)
        super(default_name, topic, registry_uri)

        @scheduled_events = []
        @semaphore = Mutex.new

        @queue_uri = queue_uri

        @interval      = interval
        @state_store   = state_store
      end

      def run
        @scheduler = Thread.new do
          Thread.current.name = 'TRMT/publisher'
          producer = Producer.new(@queue_uri)
          loop do
            @semaphore.synchronize do
              publish_overdue_events(producer)
            end
            sleep(@interval)
          end
        end

        super

        @semaphore.synchronize do
          @scheduler.terminate
        end
      end

      handle :future_event_scheduled do |event|
        @semaphore.synchronize do
          @scheduled_events << event
        end
      end

      private

      def restore_state(_sequence)
        if File.exist? @state_store
          json_state = File.read(@state_store)
          return if json_state.strip.empty?
          stored_events = JSON.parse(json_state, symbolize_names: true)
          return if stored_events.nil?
          restored_events = stored_events.map { |event| Event.from(event) }
          @semaphore.synchronize do
            @scheduled_events += restored_events
          end
        end
      end

      def persist_state(_event)
        write_state_store
      end

      def write_state_store
        File.write(@state_store, @scheduled_events.to_json)
      end

      def finalize
        # we already persisted the state... nothing to do!
      end

      def publish_overdue_events(producer)
        wall_clock_time = TrRMIte.wall_clock_time

        overdue_events = @scheduled_events.find_all { |event|
          event[:publish_at] <= wall_clock_time
        }

        overdue_events.each do |event|
          producer.publish_scheduled(event)
          @scheduled_events.delete(event)
        end

        write_state_store if overdue_events.any?
      end

    end
  end
end
